package ad1.ss16.pa;

import java.util.*;

@SuppressWarnings("unused")
public class Network {

	// Sets Java:
	// http://stackoverflow.com/questions/559839/big-o-summary-for-java-collections-framework-implementations
	private ArrayList<HashSet<Integer>> adjList = null;
	/*
	 * ArrayList: get O(1), add O(1) HashSet: add O(1), contains O(1), next
	 * O(h/n)
	 */

	private int nodes = 0;
	private int connections = 0;
	private int components = 0;

	// Wenn add oder delete erfolgreich war = false, wenn es nicht erfolgreich
	// war (v==w) passiert nichts mit isComponentsKnown
	private boolean isComponentsKnown = false;

	// Welche Node besucht wurde und welche nicht. Hat nur 1/8 der groesse von
	// einem Boolean Array
	private BitSet discovered = new BitSet();

	// true = es gibt einen Kreis | false = es gibt keinen Kreis | null =
	// Kreiszustand ist unbekannt.
	private BitSet isCycle = new BitSet();

	// fuer critical
	private int time = 0;

	// HashSet hatte einen Buffer Overflow erzeugt.
	private LinkedList<Integer> CriticalNodesList;

	// Globale Arrays fuer Critical-Suche. Es entstand sonst nen Buffer Overflow
	private int disc[];
	private int low[];
	private int parent[];
	private int child[];

	public Network(int n) {
		// am Anfang sind alle Nodes unverbunden
		this.components = n;
		this.isComponentsKnown = true;
		this.nodes = n;
		resetDiscovered();

		// es gibt keine Kreise wenn es keine Kanten gibt
		this.isCycle.set(0, false);
		this.CriticalNodesList = new LinkedList<Integer>();

		/*
		 * Zuerst wird die ArrayList erzeugt. Danach werden einzelne HashSets
		 * eingefuegt. Anfangs wurde alles in einem Schritt gemacht, aber Java
		 * wies dann viel zu viele LinkedLists zu. Er wies sie in 2^n Schritten
		 * zu, welches viel zu viel Speicher verbrauchte. Wenn man es so macht
		 * wie darunter werden genau nur soviele Arrays angelegt wie benoetigt
		 * werden. Magic
		 */
		this.adjList = new ArrayList<HashSet<Integer>>(); // Erstelle die neue
															// adjListe
		// Mache eine 2te temporaere
		ArrayList<HashSet<Integer>> adjList2 = new ArrayList<HashSet<Integer>>(n);
		for (int i = 0; i < n; i++) { // Fuege ueberall nen HashSet ein
			adjList2.add(new HashSet<Integer>());
		}
		// Kopiere die neu aufgebaute Liste in die zu nutzende.
		this.adjList = adjList2;
	}

	public int numberOfNodes() {
		return this.nodes;
	}

	public int numberOfConnections() {
		return this.connections;
	}

	public void addConnection(int v, int w) {
		// O(1) wegen HashSet bei contains

		// Schleifen und ungueltige Werte werden ignoriert
		if (v == w || v >= this.nodes || w >= this.nodes || w < 0 || v < 0) {
			return;
		}
		// Wenn v->w keine Verbindung hat
		// get O(1) + contains O(1) = 1
		if (!this.adjList.get(v).contains(w)) {

			// fuege sie hinzu
			// add O(1)
			this.adjList.get(v).add(w);
			this.adjList.get(w).add(v);

			// die Komponenten sind nun nicht mehr bekannt
			this.isComponentsKnown = false;

			// Erhoehe die connections-Anzahl
			this.connections++;

			// Es ist nicht mehr sicher ob es KEINEN Kreis gibt.
			if (this.isCycle.get(0) == false) { // Weis man ob es fix keinen
												// Kreis gibt?
				this.isCycle.clear(); // Wenn es keinen gibt, setzte es auf
										// Unbekannt.
			}
		}
	}

	public void addAllConnections(int v) {
		// O(n)
		// Verbinde die node mit allen anderen
		for (int i = 0; i < this.nodes; i++) {
			this.addConnection(v, i); // O(1)
		}

		// Einer ist mit allen Verbunden, also gibt es nur 1 Komponente
		this.components = 1;
		this.isComponentsKnown = true;
	}

	public void deleteConnection(int v, int w) {
		// O = wie addConnection

		if (this.adjList.get(v).contains(w)) { // Wenn v->w eine Verbindung hat
			this.adjList.get(v).remove(w); // entferne ihn
			this.adjList.get(w).remove(v); // entferne ihn

			this.isComponentsKnown = false;
			this.connections--;

			// Es ist nicht mehr sicher ob es EINEN Kreis gibt.
			if (this.isCycle.get(0) == true) {
				this.isCycle.clear();
			}
		}
	}

	public void deleteAllConnections(int v) {
		// O(n)

		// Schau nach welche Nodes der zu loeschende sieht und Loesche nur in
		// diesen den Node
		for (int i : this.adjList.get(v)) {
			// Loesche in jeder Node den v raus und dec. den zaehler weil eine
			// Verbindung weniger ist.
			this.adjList.get(i).remove(v);
			this.connections--;
		}

		if (!this.adjList.get(v).isEmpty()) { // Leere nun den Node selber
			this.adjList.get(v).clear();

			// Die Komponenten koennten sich nun geaendert haben
			this.isComponentsKnown = false;

			// Es ist nicht mehr sicher ob es EINEN Kreis gibt.
			this.isCycle.clear();
		}
	}

	public int numberOfComponents() {
		// DFSNUM von den Folien
		// O(n+m)
		// https://www.dropbox.com/s/dojko5w7lbk84dy/AD1_03_Graphen.pdf?dl=0
		// Seite 43

		if (isComponentsKnown == false) {
			resetDiscovered();
			this.components = 0;
			// fuer jeden Knoten v von allen Nodes
			for (int v = 0; v < this.nodes; v++) {

				// Wenn der Knoten noch nicht besucht wurde, gehe alle weiteren
				// Knoten von ihm aus ab und markiere sie. Falls der naechste
				// Knoten nicht besucht wurde ist er eine weitere Komponente
				if (this.discovered.get(v) == false) {
					this.components++;
					DFSco(v);
				}
			}
		}
		return this.components;
	}

	void DFSco(int u) {
		// O(m+n)
		this.discovered.set(u, true);
		// gehe jeden Node von der Node aus ab Wenn der Node noch nicht besucht
		// wurde, gehe eine Ebene tiefer
		for (int temp : adjList.get(u)) {
			if (!this.discovered.get(temp)) {
				DFSco(temp);
			}
		}
	}

	public boolean hasCycle() {
		// O(n*(m+n))
		// I think that depth first search solves it. If an unexplored edge
		// leads to a node visited before, then the graph contains a cycle.

		// Wenn klar ist, ob es einen Kreis gibt, gib ihn aus
		if (!this.isCycle.isEmpty()) {
			return this.isCycle.get(0);

		} else { // Wenn nicht, schau nach
			resetDiscovered();

			// fuer jeden Knoten v von allen Nodes
			// O(n)
			for (int v = 0; v < this.nodes; v++) {
				// Geh nur die Nachbarn ab die noch nicht besucht wurden
				if (this.discovered.get(v) == false) {

					// Gibt es einen Kreis mit dem aktuellen Node?
					// O(n+m)
					DFScy(v, v);

					// Wurde ein Kreis gefunden, beende die Suche
					if (this.isCycle.get(0) == true) {
						break;
					}
				} // EndIf
			} // EndFor

			return this.isCycle.get(0);
		} // EndElse
	}

	void DFScy(int u, int parent) {
		// parent wird benoetigt, damit er nicht meint das es ein Kreis ist wenn
		// er seinen Elternknoten sieht
		// O(n+m)

		// Markiere den Node als gefunden
		this.discovered.set(u, true);

		// gehe jeden Node von der Node aus ab
		for (int temp : adjList.get(u)) {
			// Wenn der Node noch nicht besucht wurde UND es nicht sein
			// Elternknoten ist, gehe eine Ebene tiefer ...
			if (!this.discovered.get(temp) && temp != parent) {
				DFScy(temp, u);

				// Wenn ein oder kein Kreis gefunden wurde, breche die Suche ab
				if (!this.isCycle.isEmpty()) {
					return;
				}

				// ... oder wenn ein Nachbar schon besucht wurde UND es ist
				// nicht der Elternknoten, dann ist es ein Kreis.
			} else if (this.discovered.get(temp) && temp != parent) {
				this.isCycle.set(0, true);
				return;
			}
		}
		return;
	}

	/*
	 * Tiefenbestimmung von BFS: Stockwerk = tiefe und Raum = node (wie bei
	 * einem Dungeon Keller) Anfangs wird immer ein einzelner Root angenommen
	 * und bestimmt damit das aktuelle Stockwerk nur 1 Raum hat. Sobald auf die
	 * Queue zugegriffen wird, wird auf einen Raum zugegriffen und der
	 * Raumzoehler -1 wird 0. Wenn der Raumzoehler 0 ist wird die aktuelle
	 * Anzahl an Queue Eintroegen gespeichert. Nun wissen wir die Elemente bzw.
	 * Anzahl der Roeume des noechsten Stockes und es wird wieder jedesmal bei
	 * einem neuen Queue Zugriff auf einen Raum der Raumzoehler um 1 verringert.
	 * Wenn wieder alle Roeume des aktuellen Stockwerkes abgegangen sind, wissen
	 * wir das wir alle Roeume im aktuellen abgegangen sind und erhoehen den
	 * distance um 1.
	 * 
	 * tl;dr; :Dungeon Crawling mit ner menge Stiegen und man laesst keinen Raum
	 * aus
	 */
	public int minimalNumberOfConnections(int start, int end) {
		// BFS mit Queue aus Folien
		// O(n+m)
		// https://www.dropbox.com/s/dojko5w7lbk84dy/AD1_03_Graphen.pdf?dl=0
		// Seite 30

		// Schleifen abfangen
		if (start == end) {
			return 0;
		}

		Queue<Integer> queue = new LinkedList<Integer>();
		// DIe Anfangsentfernung ist 0 bzw. das Erdgeschoss
		int distance = 0;
		// Am Anfang gibt es nur einen Raum, den root.
		int rooms = 1;

		resetDiscovered();

		// start node wurde besucht
		this.discovered.set(start, true);

		// Fuege Startnode zur Queue hinzu
		queue.add(start);

		// Sind noch abzugehende Raeume/Nodes da?
		while (!queue.isEmpty()) {
			// Ein Raum weniger zu besuchen bzw. der aktuelle Raum wird gerade
			// besucht
			rooms--;

			// Wenn es keine Raeume mehr im Stockwerk gibt, gehe eine Ebene
			// tiefer.
			if (rooms <= 0) {
				distance++;
				// Merke dir wieviele Raeume/Nodes der naechste Stock hat
				rooms = queue.size();
			}

			// entnehme vordersten von der queue
			// pool/remove = O(1)
			int r = queue.remove();

			// besuche zuerst die naeheren Nachbarn
			for (int temp : adjList.get(r)) {
				// Es wurde das Ende gefunden
				if (temp == end) {
					return distance;
				}

				// Wenn der Node noch nicht besucht wurde, fuege ihn zur Queue
				// hinzu
				if (!this.discovered.get(temp)) {
					// Queue add/offer = O(1)
					this.discovered.set(temp, true);
					queue.add(temp);
				}

				// Er zaehlt die Anzahl der Elemente in der Queue wenn die Ebene
				// 0 ist.
				// Danach wird solange der Zaehler nicht 0 ist immer zaehler--
				// gemacht bei jedem Queue zugriff.
			}
		} // EndWhile
		return -1; // Es wurde kein Weg zu end gefunden
	}

	/*
	 * tarjan Algo O(E + V) Zeit | O(n) Speicher
	 */
	public List<Integer> criticalNodes() {
		List<Integer> critical = new LinkedList<Integer>();

		// Initialisiere distance, low, parent, children und time
		this.CriticalNodesList.clear();
		resetDiscovered();
		this.disc = new int[nodes];
		this.low = new int[nodes];
		this.parent = new int[nodes];
		this.child = new int[nodes];
		this.time = 0;

		for (int i = 0; i < nodes; i++) {
			this.parent[i] = -1;
			this.low[i] = 0;
			this.disc[i] = 0;
			this.child[i] = 0;
		}

		// Schau in allen unbesuchten Nodes nach ob Kritische Knoten vorhanden
		// sind
		for (int i = 0; i < nodes; i++) {
			if (this.discovered.get(i) == false) {
				CriticalDfs(i);
			}
		}
		// Leere den Speicher
		this.disc = null;
		this.low = null;
		this.parent = null;
		this.child = null;

		// Fuege die nun gefundenen Werte in eine HashSet ein, weil diese keine
		// doppelten Werte erlaubt
		HashSet<Integer> temp = new HashSet<Integer>();
		temp.addAll(CriticalNodesList);
		CriticalNodesList.clear();

		critical.addAll(temp);

		return critical;
	}

	void CriticalDfs(int v) {
		// Setzte die Kinder auf 0
		this.child[v] = 0;

		// Markiere den Node als besucht
		this.discovered.set(v, true);

		// Die Entfernung und der LOW Wert ist die aktuelle Tiefe
		this.disc[v] = this.low[v] = ++this.time;

		// Gehe nun alle Nachbarknoten von v ab
		for (int x : this.adjList.get(v)) {

			// Wenn der Knoten noch nicht besucht wurde
			if (this.discovered.get(x) == false) {

				// Erhoehe seine Anzahl an Kinder
				this.child[v]++;

				// Merke dir den Elternknoten
				this.parent[x] = v;

				// Gehe eine Ebene tiefer
				CriticalDfs(x); // 0010 StackFehler?

				// Speichere dir nun die niedrigere Entfernung von v und den
				// Nachbarsknoten
				this.low[v] = Math.min(this.low[v], this.low[x]);

				// Wenn v root ist und er hat mehr als 2 Kinder
				if (this.parent[v] == -1 && this.child[v] >= 2) {
					// Fuege den Critical Node hinzu
					this.CriticalNodesList.add(v);
				}

				// Wenn er kein root ist und sein Nachbar hat ne groessere
				// Entfernung als v
				if (this.parent[v] != -1 && this.low[x] >= this.disc[v]) {
					// Fuege den Node hinzu
					this.CriticalNodesList.add(v);
				}

				// Wenn der Knoten schon besucht wurde und der Nachbarsknoten
				// ist nicht der Elternknoten
			} else if (x != this.parent[v]) {
				// Speichere die niedrigste Entfernung von v und seinem
				// Nachbarn.
				this.low[v] = Math.min(this.low[v], this.disc[x]);
			} // EndElseIf
		} // EndFor
	}

	void resetDiscovered() {
		this.discovered.clear(); // setzte discovery Liste zurueck
	}
}
